@@ -4,17 +4,19 @@ import datetime
 import inspect
 import re
 import typing
+from itertools import chain
+from operator import attrgetter
 from pathlib import Path, PurePosixPath
 from typing import Any, Callable, Dict, List, Literal, Optional, Set, Tuple, Type, Union
 from uuid import uuid4
 
 import yaml
 from ops import pebble
-from ops.model import SecretRotate
+from ops.model import SecretRotate, StatusBase
 
 from scenario.logger import logger as scenario_logger
 from scenario.mocking import _MockFileSystem, _MockStorageMount
-from scenario.runtime import trigger
+from scenario.runtime import trigger as _runtime_trigger
 
 if typing.TYPE_CHECKING:
     try:
@@ -23,6 +25,8 @@ if typing.TYPE_CHECKING:
         from typing_extensions import Self
     from ops.testing import CharmType
 
+    PathLike = Union[str, Path]
+
 logger = scenario_logger.getChild("structs")
 
 ATTACH_ALL_STORAGES = "ATTACH_ALL_STORAGES"
@@ -204,6 +208,7 @@ def _random_model_name():
 class Model(_DCBase):
     name: str = _random_model_name()
     uuid: str = str(uuid4())
+    type: Literal["kubernetes", "lxd"] = "kubernetes"
 
 
 # for now, proc mock allows you to map one command to one mocked output.
@@ -244,7 +249,17 @@ class Mount(_DCBase):
 class Container(_DCBase):
     name: str
     can_connect: bool = False
+
+    # This is the base plan. On top of it, one can add layers.
+    # We need to model pebble in this way because it's impossible to retrieve the layers from pebble
+    # or derive them from the resulting plan (which one CAN get from pebble).
+    # So if we are instantiating Container by fetching info from a 'live' charm, the 'layers' will be unknown.
+    # all that we can know is the resulting plan (the 'computed plan').
+    _base_plan: dict = dataclasses.field(default_factory=dict)
+    # We expect most of the user-facing testing to be covered by this 'layers' attribute,
+    # as all will be known when unit-testing.
     layers: Dict[str, pebble.Layer] = dataclasses.field(default_factory=dict)
+
     service_status: Dict[str, pebble.ServiceStatus] = dataclasses.field(
         default_factory=dict
     )
@@ -267,6 +282,54 @@ class Container(_DCBase):
 
     exec_mock: _ExecMock = dataclasses.field(default_factory=dict)
 
+    def _render_services(self):
+        # copied over from ops.testing._TestingPebbleClient._render_services()
+        services = {}  # type: Dict[str, pebble.Service]
+        for key in sorted(self.layers.keys()):
+            layer = self.layers[key]
+            for name, service in layer.services.items():
+                services[name] = service
+        return services
+
+    @property
+    def plan(self) -> pebble.Plan:
+        """This is the 'computed' pebble plan; i.e. the base plan plus the layers that have been added on top.
+
+        You should run your assertions on the plan, not so much on the layers, as those are input data.
+        """
+
+        # copied over from ops.testing._TestingPebbleClient.get_plan().
+        plan = pebble.Plan(yaml.safe_dump(self._base_plan))
+        services = self._render_services()
+        if not services:
+            return plan
+        for name in sorted(services.keys()):
+            plan.services[name] = services[name]
+        return plan
+
+    @property
+    def services(self) -> Dict[str, pebble.ServiceInfo]:
+        services = self._render_services()
+        infos = {}  # type: Dict[str, pebble.ServiceInfo]
+        names = sorted(services.keys())
+        for name in names:
+            try:
+                service = services[name]
+            except KeyError:
+                # in pebble, it just returns "nothing matched" if there are 0 matches,
+                # but it ignores services it doesn't recognize
+                continue
+            status = self.service_status.get(name, pebble.ServiceStatus.INACTIVE)
+            if service.startup == "":
+                startup = pebble.ServiceStartup.DISABLED
+            else:
+                startup = pebble.ServiceStartup(service.startup)
+            info = pebble.ServiceInfo(
+                name, startup=startup, current=pebble.ServiceStatus(status)
+            )
+            infos[name] = info
+        return infos
+
     @property
     def filesystem(self) -> _MockFileSystem:
         mounts = {
@@ -291,42 +354,39 @@ class Address(_DCBase):
     hostname: str
     value: str
     cidr: str
+    address: str = ""  # legacy
 
 
 @dataclasses.dataclass
 class BindAddress(_DCBase):
-    mac_address: str
     interface_name: str
-    interfacename: str  # noqa legacy
     addresses: List[Address]
+    mac_address: Optional[str] = None
 
     def hook_tool_output_fmt(self):
         # dumps itself to dict in the same format the hook tool would
-        return {
-            "bind-addresses": self.mac_address,
+        # todo support for legacy (deprecated `interfacename` and `macaddress` fields?
+        dct = {
             "interface-name": self.interface_name,
-            "interfacename": self.interfacename,
             "addresses": [dataclasses.asdict(addr) for addr in self.addresses],
         }
+        if self.mac_address:
+            dct["mac-address"] = self.mac_address
+        return dct
 
 
 @dataclasses.dataclass
 class Network(_DCBase):
     name: str
-    bind_id: int
 
     bind_addresses: List[BindAddress]
-    bind_address: str
-    egress_subnets: List[str]
     ingress_addresses: List[str]
-
-    is_default: bool = False
+    egress_subnets: List[str]
 
     def hook_tool_output_fmt(self):
         # dumps itself to dict in the same format the hook tool would
         return {
             "bind-addresses": [ba.hook_tool_output_fmt() for ba in self.bind_addresses],
-            "bind-address": self.bind_address,
             "egress-subnets": self.egress_subnets,
             "ingress-addresses": self.ingress_addresses,
         }
@@ -335,41 +395,88 @@ class Network(_DCBase):
     def default(
         cls,
         name,
-        bind_id,
         private_address: str = "1.1.1.1",
-        mac_address: str = "",
         hostname: str = "",
         cidr: str = "",
         interface_name: str = "",
+        mac_address: Optional[str] = None,
         egress_subnets=("1.1.1.2/32",),
         ingress_addresses=("1.1.1.2",),
     ) -> "Network":
         """Helper to create a minimal, heavily defaulted Network."""
         return cls(
             name=name,
-            bind_id=bind_id,
             bind_addresses=[
                 BindAddress(
-                    mac_address=mac_address,
                     interface_name=interface_name,
-                    interfacename=interface_name,
+                    mac_address=mac_address,
                     addresses=[
                         Address(hostname=hostname, value=private_address, cidr=cidr)
                     ],
                 )
             ],
-            bind_address=private_address,
             egress_subnets=list(egress_subnets),
             ingress_addresses=list(ingress_addresses),
         )
 
 
+@dataclasses.dataclass
+class _EntityStatus(_DCBase):
+    """This class represents StatusBase and should not be interacted with directly."""
+
+    # Why not use StatusBase directly? Because that's not json-serializable.
+
+    name: str
+    message: str = ""
+
+    def __eq__(self, other):
+        if isinstance(other, Tuple):
+            logger.warning(
+                "Comparing Status with Tuples is deprecated and will be removed soon."
+            )
+            return (self.name, self.message) == other
+        if isinstance(other, StatusBase):
+            return (self.name, self.message) == (other.name, other.message)
+        logger.warning(
+            f"Comparing Status with {other} is not stable and will be forbidden soon."
+            f"Please compare with StatusBase directly."
+        )
+        return super().__eq__(other)
+
+    @classmethod
+    def _from_statusbase(cls, obj: StatusBase):
+        return _EntityStatus(obj.name, obj.message)
+
+    def __iter__(self):
+        return iter([self.name, self.message])
+
+    def __repr__(self):
+        return f"<EntityStatus name={self.name}, message={self.message}>"
+
+
 @dataclasses.dataclass
 class Status(_DCBase):
-    app: Tuple[str, str] = ("unknown", "")
-    unit: Tuple[str, str] = ("unknown", "")
+    app: _EntityStatus = _EntityStatus("unknown")
+    unit: _EntityStatus = _EntityStatus("unknown")
     app_version: str = ""
 
+    def __post_init__(self):
+        for name in ["app", "unit"]:
+            val = getattr(self, name)
+            if isinstance(val, _EntityStatus):
+                pass
+            elif isinstance(val, StatusBase):
+                setattr(self, name, _EntityStatus._from_statusbase(val))
+            elif isinstance(val, tuple):
+                logger.warning(
+                    "Initializing Status.[app/unit] with Tuple[str, str] is deprecated "
+                    "and will be removed soon. \n"
+                    f"Please pass a StatusBase instance: `StatusBase(*{val})`"
+                )
+                setattr(self, name, _EntityStatus(*val))
+            else:
+                raise TypeError(f"Invalid status.{name}: {val!r}")
+
 
 @dataclasses.dataclass
 class StoredState(_DCBase):
@@ -389,7 +496,9 @@ class StoredState(_DCBase):
 
 @dataclasses.dataclass
 class State(_DCBase):
-    config: Dict[str, Union[str, int, float, bool]] = None
+    config: Dict[str, Union[str, int, float, bool]] = dataclasses.field(
+        default_factory=dict
+    )
     relations: List[Relation] = dataclasses.field(default_factory=list)
     networks: List[Network] = dataclasses.field(default_factory=list)
     containers: List[Container] = dataclasses.field(default_factory=list)
@@ -401,7 +510,7 @@ class State(_DCBase):
 
     # meta stuff: actually belongs in event data structure.
     juju_version: str = "3.0.0"
-    unit_id: str = "0"
+    unit_id: int = 0
     app_name: str = "local"
 
     # represents the OF's event queue. These events will be emitted before the event being dispatched,
@@ -416,7 +525,7 @@ class State(_DCBase):
 
     @property
     def unit_name(self):
-        return self.app_name + "/" + self.unit_id
+        return f"{self.app_name}/{self.unit_id}"
 
     def with_can_connect(self, container_name: str, can_connect: bool):
         def replacer(container: Container):
@@ -435,7 +544,9 @@ class State(_DCBase):
             status=dataclasses.replace(self.status, unit=(status, message))
         )
 
-    def get_container(self, name) -> Container:
+    def get_container(self, container: Union[str, Container]) -> Container:
+        """Get container from this State, based on an input container or its name."""
+        name = container.name if isinstance(container, Container) else container
         try:
             return next(filter(lambda c: c.name == name, self.containers))
         except StopIteration as e:
@@ -468,9 +579,10 @@ class State(_DCBase):
         meta: Optional[Dict[str, Any]] = None,
         actions: Optional[Dict[str, Any]] = None,
         config: Optional[Dict[str, Any]] = None,
+        charm_root: Optional["PathLike"] = None,
     ):
-        """Fluent API for trigger."""
-        return trigger(
+        """Fluent API for trigger. See runtime.trigger's docstring."""
+        return _runtime_trigger(
             state=self,
             event=event,
             charm_type=charm_type,
@@ -479,8 +591,11 @@ class State(_DCBase):
             meta=meta,
             actions=actions,
             config=config,
+            charm_root=charm_root,
         )
 
+    trigger.__doc__ = _runtime_trigger.__doc__
+
 
 @dataclasses.dataclass
 class _CharmSpec(_DCBase):
@@ -491,6 +606,10 @@ class _CharmSpec(_DCBase):
     actions: Optional[Dict[str, Any]] = None
     config: Optional[Dict[str, Any]] = None
 
+    # autoloaded means: trigger() is being invoked on a 'real' charm class, living in some /src/charm.py,
+    # and the metadata files are 'real' metadata files.
+    is_autoloaded: bool = False
+
     @staticmethod
     def autoload(charm_type: Type["CharmType"]):
         charm_source_path = Path(inspect.getfile(charm_type))
@@ -510,7 +629,11 @@ class _CharmSpec(_DCBase):
             actions = yaml.safe_load(actions_path.open())
 
         return _CharmSpec(
-            charm_type=charm_type, meta=meta, actions=actions, config=config
+            charm_type=charm_type,
+            meta=meta,
+            actions=actions,
+            config=config,
+            is_autoloaded=True,
         )
 
 
